#include <stdio.h>
#include <unistd.h> // usleep()
#include <sys/io.h> // I/O stuff

#include "../avrsp.h"
#include "hwctrl.h"


#include "AT91RM9200.h"
#include "gpio_etwarm.h"

#include "gpio_etwarm.c"


#define MISO AT91_PORTA,AT91C_PIO_PA0
#define MOSI AT91_PORTA,AT91C_PIO_PA1
#define SCK AT91_PORTA,AT91C_PIO_PA2

#define RESET_TH AT91_PORTA,AT91C_PIO_PA5
#define SS_TH AT91_PORTA,AT91C_PIO_PA19

#define RESET_HOT AT91_PORTA,AT91C_PIO_PA6
#define SS_HOT AT91_PORTA,AT91C_PIO_PA3

/*----------------------------------------------------------------------
  Control Variables
----------------------------------------------------------------------*/

static WORD PortBase, PortType, PortDly;
static char str_info[100];


static struct {
	DWORD	Rp;
	DWORD	Wp;
	DWORD	Ctr;
	BYTE	Buff[PIPE_WINDOW];
} RcvrFifo;




static
void fifo_put (BYTE d)
{
 if(RcvrFifo.Ctr < sizeof(RcvrFifo.Buff)) {
  RcvrFifo.Buff[RcvrFifo.Wp++] = d;
  RcvrFifo.Ctr++;
  if(RcvrFifo.Wp >= sizeof(RcvrFifo.Buff))
   RcvrFifo.Wp = 0;
 }
}




static
BYTE fifo_get (void)
{
 BYTE d = 0xFF;
 
 
 if(RcvrFifo.Ctr > 0) {
  d = RcvrFifo.Buff[RcvrFifo.Rp++];
  RcvrFifo.Ctr--;
  if(RcvrFifo.Rp >= sizeof(RcvrFifo.Buff))
   RcvrFifo.Rp = 0;
 }
 return d;
}




/* I/O control delay */

static
void iodly(void)
{
 WORD d;
 
 for(d = PortDly; d > 0; d--)
  (void) gpio_etwarm_in(MISO);
}




/*----------------------------------------------------------------------
  Public Functions
----------------------------------------------------------------------*/


/* Search and Open configuration file */

FILE *open_cfgfile(char *filename)
{
 FILE *fp;
 
 if((fp = fopen(filename, "rt")) != NULL)
  return fp;
 		
 return NULL;
}




/* Wait for dly msec */

void delay_ms (WORD dly)
{
 usleep(dly * 1000);
}



/* Flush command pipeline for SPI bridge */

int spi_flush (void)
{
 return 0;
}



/* Initialize control port and return port status */

int open_ifport (PORTPROP *pc)
{
 gpio_etwarm_init();
 
 gpio_etwarm_cfg(AT91_PORTA, 
                 AT91C_PIO_PA1 |
                 AT91C_PIO_PA2 |
		 AT91C_PIO_PA3 |
		 AT91C_PIO_PA5 |
		 AT91C_PIO_PA6 |
		 AT91C_PIO_PA19,
		 AT91_PORT_OUTPUT);
		 
 gpio_etwarm_cfg(AT91_PORTA, 
		 AT91C_PIO_PA0,
		 AT91_PORT_INPUT);
 
 
 if((pc->PortNum < 1) || (pc->PortNum > 2)) {
  sprintf(str_info, "No AVR at this port.\n");
  pc->Info1 = str_info;
  return 1;
 }
  
 PortBase = pc->PortNum;
 PortDly = pc->Delay;
 
// if(PortBase == 1) gpio_etwarm_out(SS_TH, 0xffff);
// else gpio_etwarm_out(SS_HOT, 0xffff);

gpio_etwarm_out(SS_TH, 0xffff);
gpio_etwarm_out(SS_HOT, 0xffff);
 
 return 0;
}



/* Close control port */

void close_ifport (void)
{
 if(PortBase) {
  if(PortBase == 1) gpio_etwarm_out(RESET_TH, 0xffff);
  else gpio_etwarm_out(RESET_HOT, 0xffff);
gpio_etwarm_out(SS_TH, 0);
gpio_etwarm_out(SS_HOT, 0);
  gpio_etwarm_exit();
  PortType = 0;
 }
}



/* Apply a reset sequence */

void spi_reset (void)
{
 if(PortBase == 1) gpio_etwarm_out(RESET_TH, 0);
 else gpio_etwarm_out(RESET_HOT, 0);
 delay_ms(10);
 if(PortBase == 1) gpio_etwarm_out(RESET_TH, 0xffff);
 else gpio_etwarm_out(RESET_HOT, 0xffff);
 iodly(); iodly();
 if(PortBase == 1) gpio_etwarm_out(RESET_TH, 0);
 else gpio_etwarm_out(RESET_HOT, 0);
 
 RcvrFifo.Rp = 0;	/* Flush FIFO */
 RcvrFifo.Wp = 0;
 RcvrFifo.Ctr = 0;
 delay_ms(30);		/* 30msec */
}




/* Apply an SCK pulse (for command synchronization) */

void spi_clk (void)
{
 iodly();
 gpio_etwarm_out(SCK, 0xffff);
 iodly();
 gpio_etwarm_out(SCK, 0);
}



/* Send a byte to the device */

void spi_xmit (BYTE td)
{
 int n = 8;
 
 do {
  gpio_etwarm_out(SCK, 0);			// SCK = 0
  if(td & 0x80) gpio_etwarm_out(MOSI, 0xffff);	// MOSI = Data
  else gpio_etwarm_out(MOSI, 0);
  iodly();					/* delay */
  gpio_etwarm_out(SCK, 0xffff);			// SCK = 1
  iodly();					/* delay */
  td <<= 1;
 } while (--n);
 gpio_etwarm_out(SCK, 0);			// SCK = 0
}



/* Send zero and Receive a byte from device */

BYTE spi_rcvr (BYTE mode)
{
 BYTE rd = 0;
 int n = 8;

 gpio_etwarm_out(SCK, 0);			// SCK = 0
 gpio_etwarm_out(MOSI, 0);			// MOSI = 0
 iodly();
 do {
  rd <<= 1;					/* Read MISO(ACK) */
  if(gpio_etwarm_in(MISO)) rd++;		// SCK = 0) rd++;
  gpio_etwarm_out(SCK, 0xffff);			// SCK = 1
  iodly();					/* delay */
  gpio_etwarm_out(SCK, 0);			// SCK = 0
  iodly();					/* delay */
 } while (--n);
 if(mode == RM_ASYNC)
  fifo_put(rd);
  
  return(rd);
}



/* Get delayed receiving data */

void spi_delayedget (BYTE *ptr, DWORD cnt)
{
 while (cnt--)
  *ptr++ = fifo_get();		/* Get data form receiving fifo */
}



/* Read multiple bytes from device */

void read_multi (
	int src,		/* Read from.. FLASH/EEPROM/SIGNATURE */
	DWORD adr,		/* Address */
	DWORD cnt,		/* Number of bytes to read */
	BYTE *buff		/* Read buffer */
) {
 BYTE cmd;
 DWORD n, devadr;


 if (src == FLASH
     && Device->FlashSize > (128*1024)
     && (adr & 0x1FFFF) == 0) {		/* Load extended address if needed */
  spi_xmit(C_LD_ADRX);			/* Extended address command */
  spi_xmit(0);				/* 0 */
  spi_xmit((BYTE)(adr >> 17));		/* address extended */
  spi_xmit(0);				/* 0 */
 }

 for(n = 0; n < cnt; n++) {
  switch (src) {
   case FLASH :
    cmd = (BYTE)((adr & 1) ? C_RD_PRGH : C_RD_PRGL);
    devadr = adr >> 1;
    break;
   case EEPROM :
    cmd = C_RD_EEP;
    devadr = adr;
    break;
   case SIGNATURE :
    cmd = C_RD_SIG;
    devadr = adr;
    break;
   default :
    return;
  }
  spi_xmit(cmd);			/* Eead command */
  spi_xmit((BYTE)(devadr >> 8));	/* Address high */
  spi_xmit((BYTE)devadr);		/* Address low */
  spi_rcvr(RM_ASYNC);			/* Receive data in asinc mode */
  adr++;				/* Next location */
 }
 
 spi_delayedget(buff, cnt);		/* Get delayed data */
}



/* Write a page into flash memory */

void write_page (
	DWORD adr,		/* Address (must be page boundary) */
	const BYTE *wd	/* Pointer to the page data */
) {
 BYTE d;
 int n;


 /* Skip the page if all data in the page are 0xFF */
 for(n = 0, d = 0xFF; n < Device->FlashPage; n++)
  d &= wd[n];
  
 if(d == 0xFF) return;

 /* Load the page data into page buffer */
 for (n = 0; n < Device->FlashPage; n++) {
  spi_xmit((BYTE)((n & 1) ? C_WR_PRGH : C_WR_PRGL));
  spi_xmit((BYTE)((adr + n) >> 9));
  spi_xmit((BYTE)((adr + n) >> 1));
  spi_xmit(wd[n]);
 }
 
 /* Load extended address if needed */
 if(Device->FlashSize > (128*1024)) {
  spi_xmit(C_LD_ADRX);
  spi_xmit(0);
  spi_xmit((BYTE)(adr >> 17));
  spi_xmit(0);
 }

 /* Start page programming */
 spi_xmit(C_WR_PAGE);
 spi_xmit((BYTE)(adr >> 9));
 spi_xmit((BYTE)(adr >> 1));
 spi_xmit(0);

 delay_ms(Device->FlashWait);	/* Wait for page write time */
}
