/*-----------------------------------------------------------------------*/
/*  Hardware control functions for AVRSP  R0.43b                         */
/*-----------------------------------------------------------------------*/

#include <stdio.h>
#include <unistd.h> // usleep()
#include <sys/io.h> // I/O stuff

#include "../avrsp.h"
#include "hwctrl.h"


#define _outp(a,b) outb(b,a)
#define _inp(a) inb(a)

#define LPT_DAT	((WORD)(PortBase + L_DAT))
#define LPT_STA	((WORD)(PortBase + L_STA))
#define LPT_CTL	((WORD)(PortBase + L_CTL))

/*----------------------------------------------------------------------
  Control Variables
----------------------------------------------------------------------*/

static WORD PortBase, PortType, PortDly;
static char str_info[100];

static struct {
	DWORD	Rp;
	DWORD	Wp;
	DWORD	Ctr;
	BYTE	Buff[PIPE_WINDOW];
} RcvrFifo;




/*----------------------------------------------------------------------
  Module Private Functions
----------------------------------------------------------------------*/


/*
static
int check_comport(void) {
 return 0;
}
*/




static
int check_lptport(void) {
 _outp(LPT_CTL, 0);
 _outp(LPT_DAT, 0x08);
 if((BYTE)_inp(LPT_DAT) != 0x08) return (-1);
 _outp(LPT_DAT, 0x40);
 if((BYTE)_inp(LPT_DAT) != 0x40) return (-1);
 return (0);
}




/* SPI delayed read FIFO */

static
void fifo_put (BYTE d)
{
 if(RcvrFifo.Ctr < sizeof(RcvrFifo.Buff)) {
  RcvrFifo.Buff[RcvrFifo.Wp++] = d;
  RcvrFifo.Ctr++;
  if(RcvrFifo.Wp >= sizeof(RcvrFifo.Buff))
   RcvrFifo.Wp = 0;
 }
}




static
BYTE fifo_get (void)
{
 BYTE d = 0xFF;
 
 
 if(RcvrFifo.Ctr > 0) {
  d = RcvrFifo.Buff[RcvrFifo.Rp++];
  RcvrFifo.Ctr--;
  if(RcvrFifo.Rp >= sizeof(RcvrFifo.Buff))
   RcvrFifo.Rp = 0;
 }
 return d;
}




/* I/O control delay */

static
void iodly(void)
{
 WORD d;
 
 for(d = PortDly; d > 0; d--)
  (void) _inp(PortBase);
}




/*----------------------------------------------------------------------
  Public Functions
----------------------------------------------------------------------*/


/* Search and Open configuration file */

FILE *open_cfgfile(char *filename)
{
 FILE *fp;
 
 if((fp = fopen(filename, "rt")) != NULL)
  return fp;
 		
 return NULL;
}




/* Wait for dly msec */

void delay_ms (WORD dly)
{
 usleep(dly * 1000);
}



/* Flush command pipeline for SPI bridge */

int spi_flush (void)
{
 return 0;
}



/* Initialize control port and return port status */

int open_ifport (PORTPROP *pc)
{
 static const WORD PortsLpt[] = { LPT1ADR, LPT2ADR, LPT3ADR };

 if(pc->PortClass != TY_LPT) {
  pc->Info1 = "At the moment only LPT interfaces are supported in linux.\n";
  return 1;
 }

 if(iopl(3) == -1) {
  pc->Info1 = "Unable to allow I/O.\n";
  return 1;
 }
 
 PortDly = pc->Delay;
 
 /* Use LPT port in direct I/O */
 if(pc->PortClass == TY_LPT) {
 
  if(pc->PortNum < 100){
   if((pc->PortNum < 1)||(pc->PortNum > 3)) {
    pc->Info1 = "Invalid Port#.\n";
     return 1;
   }
   PortBase = PortsLpt[pc->PortNum - 1];
  } else {
   PortBase = pc->PortNum;
  }

  if(check_lptport()) {
   sprintf(str_info, "No LPT%u(0x%X) port.\n", pc->PortNum, PortBase);
   pc->Info1 = str_info;
   return 1;
  }
  
 	_outp(LPT_DAT, 0x20);	/* Check if the adapter is ByteBlasterMV (D5-ACK) */
	if(_inp(LPT_STA) & S_ACK) {
	 _outp(LPT_DAT, 0);
	 if((_inp(LPT_STA) & S_ACK) == 0) {
		_outp(LPT_CTL, BA_ENA);
		pc->Info1 = "Altera ByteBlasterMV was found.\n";
		PortType = TY_ALTERA;
		if(_inp(LPT_STA) & S_ERR) return 0;	/* Check target power */
		pc->Info2 = "But target power is OFF.\n";
		return 1;
	 }
	}
		_outp(LPT_DAT, 0x80);	/* Check if the adapter is AVRSP (D7-PE) */
		if(_inp(LPT_STA) & S_PE) {
			_outp(LPT_DAT, 0);
			if((_inp(LPT_STA) & S_PE) == 0) {
				_outp(LPT_DAT, B_ENA);
				pc->Info1 = "AVRSP adapter was found.\n";
				PortType = TY_AVRSP;
				if(_inp(LPT_STA) & S_ERR) return 0;	/* Check target power */
				pc->Info2 = "But target power is OFF.\n";
				return 1;
			}
		}
		_outp(LPT_DAT, 0x40);	/* Check if the adapter is Xilinx JTAG (D6-BUSY-PE) */
		if((_inp(LPT_STA) & (S_PE | S_BUSY)) == S_PE) {
			_outp(LPT_DAT, 0);
			if((_inp(LPT_STA) & (S_PE | S_BUSY)) == S_BUSY) {
				pc->Info1 = "Xilinx JTAG adapter was found.\n";
				PortType = TY_XILINX;
				if(_inp(LPT_STA) & S_ERR) return 0;	/* Check target power */
				pc->Info2 = "But target power is OFF.\n";
				return 1;
			}
		}
		_outp(LPT_DAT, 0x40);	/* Check if the adapter is Lattice ISP (D6-PE) */
		if(_inp(LPT_STA) & S_PE) {
			_outp(LPT_DAT, 0);
			if((_inp(LPT_STA) & S_PE) == 0) {
				pc->Info1 = "Lattice ISP adapter was found.\n";
				PortType = TY_LATTICE;
				if(_inp(LPT_STA) & S_ERR) return 0;	/* Check target power */
				pc->Info2 = "But target power is OFF.\n";
				return 1;
			}
		}
		_outp(LPT_DAT, 0x01);	/* Check if the adapter is STK200 dongle (D0-PE) */
		if(_inp(LPT_STA) & S_PE) {
			_outp(LPT_DAT, 0);
			if((_inp(LPT_STA) & S_PE) == 0) {
				pc->Info1 = "STK200 ISP dongle was found.\n";
				PortType = TY_STK200;
				if(_inp(LPT_STA) & S_ERR) return 0;	/* Check target power */
				pc->Info2 = "But target power is OFF.\n";
				return 1;
			}
		}
//		sprintf(str_info, "No ISP adapter on the LPT%u.\n", pc->PortNum);
  sprintf(str_info, "No ISP adapter on the LPT%u(0x%X).\n", pc->PortNum, PortBase);
  pc->Info1 = str_info;
  return 1;
 }

 pc->Info1 = "Invalid port class.\n";
 return 1;
}



/* Close control port */

void close_ifport (void)
{
 switch(PortType) {
  case TY_AVRSP :
	 _outp(LPT_DAT, 0);
	 break;
	 
  case TY_STK200 :
   _outp(LPT_DAT, BS_DIS);
   break;
   
  case TY_XILINX :
	 _outp(LPT_DAT, BX_DIS1|BX_DIS2);
	 break;

	case TY_LATTICE :
	 _outp(LPT_DAT, BL_DIS);
	 break;

	case TY_ALTERA :
	 _outp(LPT_CTL, 0);
	 break;
 }
 
 PortType = 0;
}



/* Apply a reset sequence */

void spi_reset (void)
{
 switch (PortType) {
		case TY_AVRSP :
			_outp(LPT_DAT, B_ENA);
			delay_ms(10);
			_outp(LPT_DAT, B_ENA|B_RES);
			iodly(); iodly();
			_outp(LPT_DAT, B_ENA);
			break;

		case TY_STK200 :
			_outp(LPT_DAT, 0);
			delay_ms(10);
			_outp(LPT_DAT, BS_RES);
			iodly(); iodly();
			_outp(LPT_DAT, 0);
			break;

		case TY_XILINX :
			_outp(LPT_DAT, BX_DIS2);
			delay_ms(10);
			_outp(LPT_DAT, BX_DIS2|BX_RES);
			iodly(); iodly();
			_outp(LPT_DAT, BX_DIS2);
			break;

		case TY_LATTICE :
			_outp(LPT_DAT, 0);
			delay_ms(10);
			_outp(LPT_DAT, BL_RES);
			iodly(); iodly();
			_outp(LPT_DAT, 0);
			break;

		case TY_ALTERA :
			_outp(LPT_DAT, 0);
			delay_ms(10);
			_outp(LPT_DAT, BA_RES);
			iodly(); iodly();
			_outp(LPT_DAT, 0);
			break;

	}
 RcvrFifo.Rp = 0;	/* Flush FIFO */
 RcvrFifo.Wp = 0;
 RcvrFifo.Ctr = 0;
 delay_ms(30);		/* 30msec */
}




/* Apply an SCK pulse (for command synchronization) */

void spi_clk (void)
{
 switch (PortType) {

		case TY_AVRSP :
			iodly();
			_outp(LPT_DAT, B_ENA|B_SCK);
			iodly();
			_outp(LPT_DAT, B_ENA);
			break;

		case TY_STK200 :
			iodly();
			_outp(LPT_DAT, BS_SCK);
			iodly();
			_outp(LPT_DAT, 0);
			break;

		case TY_XILINX :
			iodly();
			_outp(LPT_DAT, BX_DIS2|BX_SCK);
			iodly();
			_outp(LPT_DAT, BX_DIS2);
			break;

		case TY_LATTICE :
			iodly();
			_outp(LPT_DAT, BL_SCK);
			iodly();
			_outp(LPT_DAT, 0);
			break;

		case TY_ALTERA :
			_outp(LPT_DAT, BA_SCK);
			iodly();
			_outp(LPT_DAT, 0);
			break;
	}
	
}



/* Send a byte to the device */

void spi_xmit (BYTE td)
{
 BYTE a;
 int n = 8;
 
 
 switch (PortType) {
 		case TY_AVRSP :
			do {
				a = (BYTE)((td & 0x80) ? B_ENA | B_MOSI : B_ENA);	/* MOSI(D1) = data, SCK(D0) = "L" */
				_outp(LPT_DAT, a);
				iodly();									/* delay */
				_outp(LPT_DAT, a | (BYTE)B_SCK);					/* SCK(D0) = "H" */
				iodly();									/* delay */
				td <<= 1;
			} while (--n);
			_outp(LPT_DAT, a);								/* SCK(D0) = "L" */
			break;

		case TY_STK200 :
			do {
				a = (BYTE)((td & 0x80) ? BS_MOSI : 0);	/* MOSI(D5) = data, SCK(D4) = "L" */
				_outp(LPT_DAT, a);
				iodly();							/* delay */
				_outp(LPT_DAT, a | (BYTE)BS_SCK);			/* SCK(D4) = "H" */
				iodly();							/* delay */
				td <<= 1;
			} while (--n);
			_outp(LPT_DAT, a);						/* SCK(D4) = "L" */
			break;

		case TY_XILINX :
			do {
				a = (BYTE)((td & 0x80) ? BX_DIS2 | BX_MOSI : BX_DIS2); /* MOSI(D0) = data, SCK(D1) = "L" */
				_outp(LPT_DAT, a);
				iodly();									/* delay */
				_outp(LPT_DAT, a | (BYTE)BX_SCK);					/* SCK(D1) = "H" */
				iodly();									/* delay */
				td <<= 1;
			} while (--n);
			_outp(LPT_DAT, a);								/* SCK(D1) = "L" */
			break;

		case TY_LATTICE :
			do {
				a = (BYTE)((td & 0x80) ? BL_MOSI : 0);	/* MOSI(D0) = data, SCK(D1) = "L" */
				_outp(LPT_DAT, a);
				iodly();							/* delay */
				_outp(LPT_DAT, a | (BYTE)BL_SCK);			/* SCK(D1) = "H" */
				iodly();							/* delay */
				td <<= 1;
			} while (--n);
			_outp(LPT_DAT, a);						/* SCK(D1) = "L" */
			break;

		case TY_ALTERA :
			do {
				a = (BYTE)((td & 0x80) ? BA_MOSI : 0);	/* MOSI(D6) = data, SCK(D0) = "L" */
				_outp(LPT_DAT, a);
				iodly();							/* delay */
				_outp(LPT_DAT, a | (BYTE)BA_SCK);			/* SCK(D0) = "H" */
				iodly();							/* delay */
				td <<= 1;
			} while (--n);
			_outp(LPT_DAT, a);						/* SCK(D0) = "L" */
			break;
	}

}



/* Send zero and Receive a byte from device */

BYTE spi_rcvr (BYTE mode)
{
 BYTE rd = 0;
 int n = 8;
 
 	switch (PortType) {
		case TY_AVRSP :
			_outp(LPT_DAT, B_ENA);					/* MOSI(D0) = "L" */
			iodly();
			do {
				rd <<= 1;							/* Read MISO(BUSY) */
				if((_inp(LPT_STA) & S_BUSY) == 0) rd++;
				_outp(LPT_DAT, B_ENA | B_SCK);		/* SCK(D1) = "H" */
				iodly();							/* delay */
				_outp(LPT_DAT, B_ENA);				/* SCK(D1) = "L" */
				iodly();							/* delay */
			} while (--n);
			if(mode == RM_ASYNC)
				fifo_put(rd);
			break;

		case TY_STK200 :
			_outp(LPT_DAT, 0);						/* MOSI(D5) = "L" */
			iodly();
			do {
				rd <<= 1;							/* Read MISO(ACK) */
				if(_inp(LPT_STA) & S_ACK) rd++;
				_outp(LPT_DAT, BS_SCK);				/* SCK(D4) = "H" */
				iodly();							/* delay */
				_outp(LPT_DAT, 0);					/* SCK(D4) = "L" */
				iodly();							/* delay */
			} while (--n);
			if(mode == RM_ASYNC)
				fifo_put(rd);
			break;

		case TY_XILINX :
			_outp(LPT_DAT, BX_DIS2);				/* MOSI(D0) = "L" */
			iodly();
			do {
				rd <<= 1;							/* Read MISO(SEL) */
				if(_inp(LPT_STA) & S_SEL) rd++;
				_outp(LPT_DAT, BX_DIS2 | BX_SCK);	/* SCK(D1) = "H" */
				iodly();							/* delay */
				_outp(LPT_DAT, BX_DIS2);			/* SCK(D1) = "L" */
				iodly();							/* delay */
			} while (--n);
			if(mode == RM_ASYNC)
				fifo_put(rd);
			break;

		case TY_LATTICE :
			_outp(LPT_DAT, 0);						/* MOSI(D0) = "L" */
			iodly();
			do {
				rd <<= 1;							/* Read MISO(ACK) */
				if(_inp(LPT_STA) & S_ACK) rd++;
				_outp(LPT_DAT, BL_SCK);				/* SCK(D1) = "H" */
				iodly();							/* delay */
				_outp(LPT_DAT, 0);					/* SCK(D1) = "L" */
				iodly();							/* delay */
			} while (--n);
			if(mode == RM_ASYNC)
				fifo_put(rd);
			break;

		case TY_ALTERA :
			_outp(LPT_DAT, 0);						/* MOSI(D6) = "L" */
			iodly();
			do {
				rd <<= 1;							/* Read MISO(BUSY) */
				if((_inp(LPT_STA) & S_BUSY) == 0) rd++;
				_outp(LPT_DAT, BA_SCK);				/* SCK(D0) = "H" */
				iodly();							/* delay */
				_outp(LPT_DAT, 0);					/* SCK(D0) = "L" */
				iodly();							/* delay */
			} while (--n);
			if(mode == RM_ASYNC)
				fifo_put(rd);
			break;
	}

  return(rd);
}



/* Get delayed receiving data */

void spi_delayedget (BYTE *ptr, DWORD cnt)
{
 while (cnt--)
  *ptr++ = fifo_get();		/* Get data form receiving fifo */
}



/* Read multiple bytes from device */

void read_multi (
	int src,		/* Read from.. FLASH/EEPROM/SIGNATURE */
	DWORD adr,		/* Address */
	DWORD cnt,		/* Number of bytes to read */
	BYTE *buff		/* Read buffer */
) {
 BYTE cmd;
 DWORD n, devadr;


 if (src == FLASH
     && Device->FlashSize > (128*1024)
     && (adr & 0x1FFFF) == 0) {		/* Load extended address if needed */
  spi_xmit(C_LD_ADRX);			/* Extended address command */
  spi_xmit(0);				/* 0 */
  spi_xmit((BYTE)(adr >> 17));		/* address extended */
  spi_xmit(0);				/* 0 */
 }

 for(n = 0; n < cnt; n++) {
  switch (src) {
   case FLASH :
    cmd = (BYTE)((adr & 1) ? C_RD_PRGH : C_RD_PRGL);
    devadr = adr >> 1;
    break;
   case EEPROM :
    cmd = C_RD_EEP;
    devadr = adr;
    break;
   case SIGNATURE :
    cmd = C_RD_SIG;
    devadr = adr;
    break;
   default :
    return;
  }
  spi_xmit(cmd);			/* Eead command */
  spi_xmit((BYTE)(devadr >> 8));	/* Address high */
  spi_xmit((BYTE)devadr);		/* Address low */
  spi_rcvr(RM_ASYNC);			/* Receive data in asinc mode */
  adr++;				/* Next location */
 }
 
 spi_delayedget(buff, cnt);		/* Get delayed data */
}



/* Write a page into flash memory */

void write_page (
	DWORD adr,		/* Address (must be page boundary) */
	const BYTE *wd	/* Pointer to the page data */
) {
 BYTE d;
 int n;


 /* Skip the page if all data in the page are 0xFF */
 for(n = 0, d = 0xFF; n < Device->FlashPage; n++)
  d &= wd[n];
 if(d == 0xFF) return;

 /* Load the page data into page buffer */
 for (n = 0; n < Device->FlashPage; n++) {
  spi_xmit((BYTE)((n & 1) ? C_WR_PRGH : C_WR_PRGL));
  spi_xmit((BYTE)((adr + n) >> 9));
  spi_xmit((BYTE)((adr + n) >> 1));
  spi_xmit(wd[n]);
 }
 
 /* Load extended address if needed */
 if(Device->FlashSize > (128*1024)) {
  spi_xmit(C_LD_ADRX);
  spi_xmit(0);
  spi_xmit((BYTE)(adr >> 17));
  spi_xmit(0);
 }

 /* Start page programming */
 spi_xmit(C_WR_PAGE);
 spi_xmit((BYTE)(adr >> 9));
 spi_xmit((BYTE)(adr >> 1));
 spi_xmit(0);

 delay_ms(Device->FlashWait);	/* Wait for page write time */
}
